from typing import Tuple, Any, List, Dict

import matplotlib.pyplot as plt

from isthmuslib import VectorMultiset
from src.isthmuslib.logging import parse_string_with_embedded_json, parse_string_with_manual_tokens
from src.isthmuslib.utils import risky_cast, process_queue, recursive_batch_evaluation, return_best_input


def test_risky_cast():
    def val_and_type(input_value: Any) -> Tuple[Any, type]:
        return type(r := risky_cast(input_value)), r

    assert (int, 5) == val_and_type('5')
    assert (float, 4.55) == val_and_type('4.55')
    assert (str, 'foo') == val_and_type('foo')
    assert (dict, {'a': 5, 'b': 6}) == val_and_type("a=5, b=6")
    assert (bool, False) == val_and_type('False')
    assert (bool, True) == val_and_type('True')
    assert (dict, {'a': 5, 'b': True}) == val_and_type("a=5, b=True")


def test_extract_json():
    s: str = """The quick
    [[EMBEDDED_JSON_LINE]]{"color": "red", "value": "#f00"}
    Brown fox jumped 
    [[EMBEDDED_JSON_LINE]]{"color": "green", "value": "#0f0"}
    [[EMBEDDED_JSON_LINE]]{"color": "blue", "value": "#00f"}
    Over the foobar
    [[EMBEDDED_JSON_LINE]]{"color": "cyan", "value": "#0ff"}
    [[EMBEDDED_JSON_LINE]]{"color": "black", "value": "#000"}
    """
    extracted = parse_string_with_embedded_json(s)
    assert extracted.iloc[1, 0] == 'green'
    assert extracted.iloc[1, 1] == '#0f0'


def foobar(x: int) -> int:
    return x * 10


def starbar(x: int, y: int) -> str:
    return f"x={x} and y={y} so {x*y=}"


def test_process_queue():
    inputs: List[int] = list(range(25))
    star_inputs: List[Tuple[int, int]] = [(x, x + 2) for x in list(range(10))]

    result = process_queue(foobar, inputs, num_workers=5, cuts_per_worker=1, batching=False)
    assert result[:3] == [0, 10, 20]
    result = process_queue(foobar, inputs, num_workers=5, cuts_per_worker=1, batching=True)
    assert result[:3] == [0, 10, 20]
    result = process_queue(starbar, star_inputs, num_workers=5, cuts_per_worker=1, batching=False)
    assert result[:3] == ['x=0 and y=2 so x*y=0', 'x=1 and y=3 so x*y=3', 'x=2 and y=4 so x*y=8']
    try:
        print(process_queue(starbar, star_inputs, num_workers=5, cuts_per_worker=1, batching=True))
    except NotImplementedError as e:
        print(f"... not implemented yet")


def test_parsing():
    log: str = """
    log-2018-08-31-20-07-29:2018-08-31 20:07:24.748	[P2P2]	INFO protocol_handler.inl:1171	[1;33m[80.241.216.213:18080 OUT]  Synced 49156/1651224 (0.282567 sec, 353.898367 blocks/sec), 99.548294 MB [49255:o][0m
    log-2018-08-31-20-07-29:2018-08-31 20:07:25.040	[P2P2]	INFO protocol_handler.inl:1171	[1;33m[80.241.216.213:18080 OUT]  Synced 49256/1651224 (0.291497 sec, 343.056704 blocks/sec), 99.046715 MB [49355:][0m
    log-2018-08-31-20-07-29:2018-08-31 20:07:25.280	[P2P2]	INFO protocol_handler.inl:1171	[1;33m[80.241.216.213:18080 OUT]  Synced 49356/1651224 (0.239763 sec, 417.078532 blocks/sec), 98.918274 MB [49455:oo][0m
    log-2018-08-31-20-07-29:2018-08-31 20:07:25.644	[P2P2]	INFO protocol_handler.inl:1171	[1;33m[80.241.216.213:18080 OUT]  Synced 49456/1651224 (0.363321 sec, 275.238701 blocks/sec), 98.767830 MB [49555:o.][0m
    log-2018-08-31-20-07-29:2018-08-31 20:07:25.981	[P2P2]	INFO protocol_handler.inl:1171	[1;33m[80.241.216.213:18080 OUT]  Synced 49556/1651224 (0.335829 sec, 297.770592 blocks/sec), 98.502403 MB [49655:...][0m
    log-2018-08-31-20-07-29:2018-08-31 20:07:26.264	[P2P2]	INFO protocol_handler.inl:1171	[1;33m[80.241.216.213:18080 OUT]  Synced 49656/1651224 (0.281597 sec, 355.117420 blocks/sec), 98.384758 MB [49755:oo....][0m
    log-2018-08-31-20-07-29:2018-08-31 20:07:26.776	[P2P2]	INFO protocol_handler.inl:1171	[1;33m[80.241.216.213:18080 OUT]  Synced 49756/1651224 (0.404512 sec, 247.211455 blocks/sec), 98.054283 MB [49855:oo.....][0m
    log-2018-08-31-20-07-29:2018-08-31 20:07:27.066	[P2P2]	INFO protocol_handler.inl:1171	[1;33m[80.241.216.213:18080 OUT]  Synced 49856/1651224 (0.289512 sec, 345.408826 blocks/sec), 98.258347 MB [49955:oo.....][0m
    """

    # Set the rules for extracting the vectors
    record_delimiter: str = 'lo'
    tokens_dictionary: Dict[str, Tuple[str, str]] = {
        # varname: (left_token, right_token)
        'date_time_stamp': ('g-', ':'),
        'height': ('Synced ', '/'),
        'time_to_load': ('(', ' sec'),
    }

    # Extract the text into a timeseries
    timeseries: VectorMultiset = VectorMultiset(
        data=parse_string_with_manual_tokens(log,
                                             record_delimiter=record_delimiter,
                                             parallelize_processing=True,
                                             tokens_dictionary=tokens_dictionary))
    print(timeseries.data)


def mock_hill(kwargs) -> float:
    x = kwargs.get('x')
    y = kwargs.get('y')
    return 1 / x + 1 / (y ** 2)


def test_recursive_batch_evaluation():
    inital_point: Dict[str, Any] = {'x': 400, 'y': 300}
    current_best_input, history = recursive_batch_evaluation(
        func=mock_hill,
        initial_input=inital_point,
        selection_method=max,
        batch_generator_kwargs={'width_prct': 10, 'num_samples': 3},
        max_deep=5,
        print_progress=True,
        num_workers=64,
        return_history=True,
    )
    assert current_best_input['x'] < inital_point['x']
    assert current_best_input['y'] < inital_point['y']


def test_selector():
    inputs: List[Any] = ['abc', 'def', 'efg']
    eval_outputs: List[Dict[str, Any]] = [{'foobar': 'xyz', 'fitness': 4}, {'fitness': 99, 'baz': 12345},
                                          {'fitness': -5}]
    best = return_best_input(inputs, eval_outputs)
    assert best == 'def'


def y(x: float) -> float:
    return 10 - (x - 5) ** 2


def y_return_scalar(d) -> float:
    return y(d.get('x'))


def test_recursive_batch_evaluation_new_with_scalar(show_plot: bool = True):
    # Set up the run
    start_at: float = 15

    best_x, run_data = recursive_batch_evaluation(
        func=y_return_scalar,
        initial_input={'x': start_at},
        max_deep=20,
        return_history=True,
        evaluate_initial_inputs=True,
        infinite_memory=True,
        selection_method=max,
        catch_exceptions=False,
        batch_generator_kwargs={
            "width_prct": 50,
            "num_samples": 5,
            "fields": ['x'],
            "width_temperature_prct": 0,
        },
    )

    print(f"{best_x=}\n{run_data}")
    if show_plot:
        x_vals: List[float] = list(range(-10, 25))
        plt.figure(facecolor='w')
        plt.plot(x_vals, [y_return_scalar({'x': x}) for x in x_vals], color='k')
        x_vec = [d.get('in').get('x') for d in run_data]
        y_vec = [x.get('out') for x in run_data]
        plt.scatter(x_vec, y_vec, c=list(range(len(x_vec))), s=50)
        plt.show()

    assert best_x['x'] < start_at


def y_return_dict(d) -> dict:
    return {'y': y(d.get('x'))}


def test_recursive_batch_evaluation_new_with_dict(show_plot: bool = True):
    # Set up the run
    start_at: float = 15

    best_x, run_data = recursive_batch_evaluation(
        func=y_return_dict,
        initial_input={'x': start_at},
        max_deep=5,
        return_history=True,
        evaluate_initial_inputs=True,
        infinite_memory=True,
        selection_method='y',
        catch_exceptions=False,
        batch_generator_kwargs={
            "width_prct": 50,
            "num_samples": 5,
            "fields": ['x'],
            "width_temperature_prct": 0,
        },
    )

    print(f"{best_x=}\n{run_data}")
    if show_plot:
        x_vals: List[float] = list(range(-10, 25))
        plt.figure(facecolor='w')
        plt.plot(x_vals, [y_return_dict({'x': x})['y'] for x in x_vals], color='k')
        x_vec = [d.get('in').get('x') for d in run_data]
        y_vec = [x.get('out').get('y') for x in run_data]
        plt.scatter(x_vec, y_vec, c=list(range(len(x_vec))), s=50)
        plt.show()

    print(f"{run_data[-1]['in']}")
    assert run_data[-1]['in']['incremented_counter'] == 4
    assert run_data[-1]['in']['clock_time_at_start'] > 100000
    assert best_x['x'] < start_at



from typing import List, Tuple, Dict, Any, Union
from matplotlib import pyplot as plt
import pandas as pd
from .config import Style
from .plotting import visualize_x_y, visualize_1d_distribution
import numpy as np


def data_diffs(array: Any) -> List[Union[float, int]]:
    """Tiny helper function that returns the differences

    :param array: the array (in any form)
    :return: spacings between the elements
    """
    return [float(x[1]) - float(x[0]) for x in zip(array, array[1:])]


def is_uniform(array: Any) -> bool:
    """Tiny helper function, checks if an array appears to be a uniformly spaced list of numbers

    :param array: the array (in any form)
    :return: True if the elements are uniformly spaced
    """
    return len(set(data_diffs(array))) == 1


def any_missing_data(array: Any) -> bool:
    """Tiny helper function, checks whether a data array appears to have any missing data, wraps a pandas util

    :param array: the array (in any form)
    :return: True if any data points appear to be missing
    """
    if isinstance(array, pd.DataFrame):
        return array.isnull().values.any()
    else:
        return pd.DataFrame({"_": array}).isnull().values.any()


def basis_quality_checks(array: Any) -> Tuple[bool, str]:
    """Checks data quality and returns both a True/False flag and a string with an explanation

    :param array: the array (in any form)
    :return: True if OK, and a string with explanation either way
    """
    if ~any_missing_data(array):
        summary_string: str = "[OK] COMPLETENESS: no apparent missing data\n"
    else:
        summary_string: str = "[WARNING] COMPLETENESS: appears to be missing data"
        return (
            False,
            f"{summary_string}\n (note: cannot check uniformity with missing data)",
        )

    if is_uniform(array):
        summary_string += f"[OK] UNIFORMITY: appears to be uniform with spacing: {data_diffs(array)[0]}\n"
        return True, summary_string
    else:
        diffs_sorted: List[float] = sorted(list(set(data_diffs(array))))
        diffs_sorted.reverse()
        summary_string += f"[WARNING] UNIFORMITY: multiple ({len(diffs_sorted)}) spacings: {diffs_sorted}\n"
        return False, summary_string


def passes_basis_quality_checks(array: Any) -> bool:
    """Super thin wrapper around basis_quality_checks that drops the explanation string

    :param array: the array (in any form)
    :return: True if OK
    """
    result, _ = basis_quality_checks(array)
    return result


def fill_ratio(array: Any) -> float:
    """Helper function that identifies the ratio of missing values in a uniformly-spaced sequence of values.
        Attempts to infer spacing from median of diffs. If used on non-uniform data the result is not meaningful.

    Use example: [10, 20, 40, 50] has 0.8 fill ratio because it is missing the 30 to complete the set

    :param array: Any 1D vector, like a list of integers or floats, numpy array, etc
    :return: fractional amount of the array
    """
    ordered: Any = sorted(array)
    theoretical_count = 1 + (max(array) - min(array)) / np.nanmedian(
        [x[1] - x[0] for x in zip(ordered, ordered[1:])]
    )
    return len(array) / theoretical_count


def basis_quality_plots(
    array: List[float], style: Style = None, which_plots: List[str] = None, **kwargs
) -> List[plt.Figure]:
    """Generates some standard plots for checking completeness and uniformity of an array

    :param array: the array (in any form)
    :param style: optional Style configuration object
    :param which_plots: which plots to make, choices include:
        - 'missing_data_scatter'
        - 'interval_histogram'
        - 'interval_scatter'
    :param kwargs: additional kwargs passed through to visualize_x_y
    :return:
    """
    # Style and init
    if not style:
        style: Style = Style()
    mappings: Dict[bool, Tuple[str, Any]] = {
        True: ("[OK]", style.good_color),
        False: ("[WARNING]", style.bad_color),
    }
    h: List[plt.Figure] = []

    # Scatter plot of missing data
    if (not which_plots) or ("missing_data_scatter" in which_plots):
        if not any_missing_data(array):
            title: str = f"{mappings[True][0]} no missing data to plot"
        else:
            title: str = f"{mappings[False][0]} missing data"
        title += f"\n(want all 0's, on the {style.good_color} bottom line)"
        vals_y = [int((x is None) or np.isnan(float(x))) for x in array]
        f: plt.Figure = visualize_x_y(
            range(len(vals_y)),
            vals_y,
            xlabel="index",
            ylabel="is_missing (1=True)",
            types="scatter",
            style=style.override({"color": "k", "markersize": 2 * style.markersize}),
            title=title,
            **kwargs,
        )
        plt.axhline(y=0, linewidth=2, linestyle=":", c=style.good_color)
        plt.axhline(y=1, linewidth=2, linestyle=":", c=style.bad_color)
        plt.ylim([-0.5, 1.5])
        xlims: Tuple[float, float] = plt.xlim()
        if "y" not in kwargs.get("log_axes", ""):
            plt.text(
                xlims[0] + 0.02 * (xlims[1] - xlims[0]),
                0.05,
                "0 -> good",
                color=style.good_color,
                size=style.label_fontsize,
            )
        plt.text(
            xlims[0] + 0.02 * (xlims[1] - xlims[0]),
            1.05,
            "1 = NaN or None -> bad",
            color=style.bad_color,
            size=style.label_fontsize,
        )
        h.append(f)

    # Histogram of data spacing
    if (not which_plots) or ("intervals_histogram" in which_plots):
        if not any_missing_data(array):
            vals: List[Any] = sorted(data_diffs(array))
            title: str = f"{mappings[is_uniform(array)][0]} data intervals distribution \n(want only 1 value)"
            color: Any = mappings[is_uniform(array)][1]
        else:
            vals: List[Any] = [[]]
            title: str = f"{mappings[False][0]} Cannot make intervals hist with missing data! Plot is blank."
            color: Any = mappings[False][1]
        h.append(
            visualize_1d_distribution(
                vals,
                xlabel="difference between elements",
                ylabel="counts",
                title=title,
                style=style.override({"color": color}),
                **kwargs,
            )
        )

    # Scatter plot of data spacing
    if (not which_plots) or ("intervals_scatter" in which_plots):
        if not any_missing_data(array):
            sorted_array: Any = sorted(array)
            vals_x: List[Any] = sorted_array[1:]
            vals_y: List[Any] = data_diffs(sorted_array)
            title: str = f"{mappings[is_uniform(array)][0]} data intervals over time\n(want 1 horizontal line)"
            color: Any = mappings[is_uniform(array)][1]
        else:
            vals_x = vals_y = [[]]
            title: str = f"{mappings[False][0]} Cannot make diffs plot with missing data! Plot is blank."
            color: Any = mappings[False][1]
        h.append(
            visualize_x_y(
                vals_x,
                vals_y,
                types="line",
                xlabel="basis",
                ylabel="difference between elements",
                title=title,
                style=style.override({"color": color}),
                **kwargs,
            )
        )

    # Give a hint if the wrong which_plots are specified
    if (not h) and (which_plots is not None):
        print(
            f"There are no plots; this might be because basis_quality_plots does not recognize {which_plots=}."
        )

    return h
